# date4/datetime.rb: Written by Tadayoshi Funaba 1998-2006
# $Id: datetime.rb,v 1.7 2006-12-30 21:49:56+09 tadf Exp $

class DateTime4 < Date4

  def self.jd(*args)
    jd, h, min, s, tz, sg, = parsearg([[0, :jd],
				       [0, :h, :hour],
				       [0, :min, :minute],
				       [0, :s, :sec, :second],
				       [0, :of, :offset, :tz, :timezone],
				       [ITALY, :start]], args)

    unless       _valid_jd?(jd, sg) &&
	   (fr = _valid_time?(h, min, s))
      raise ArgumentError, 'invalid date'
    end
    if String === tz
      tz = (zone_to_diff(tz) || 0).to_r/86400
    end
    new!([jd, fr], tz, sg)
  end

  def self.ordinal(*args)
    y, d, h, min, s, tz, sg, = parsearg([[-4712, :y, :year],
					 [1, :d, :day, :yday],
					 [0, :h, :hour],
					 [0, :min, :minute],
					 [0, :s, :sec, :second],
					 [0, :of, :offset, :tz, :timezone],
					 [ITALY, :start]], args)

    unless (jd = _valid_ordinal?(y, d, sg)) &&
	   (fr = _valid_time?(h, min, s))
      raise ArgumentError, 'invalid date'
    end
    if String === tz
      tz = (zone_to_diff(tz) || 0).to_r/86400
    end
    new!([jd, fr], tz, sg)
  end

  def self.civil(*args)
    y, m, d, h, min, s, tz, sg, = parsearg([[-4712, :y, :year],
					    [1, :m, :mon, :month],
					    [1, :d, :day, :mday],
					    [0, :h, :hour],
					    [0, :min, :minute],
					    [0, :s, :sec, :second],
					    [0, :of, :offset, :tz, :timezone],
					    [ITALY, :start]], args)

    unless (jd = _valid_civil?(y, m, d, sg)) &&
	   (fr = _valid_time?(h, min, s))
      raise ArgumentError, 'invalid date'
    end
    if String === tz
      tz = (zone_to_diff(tz) || 0).to_r/86400
    end
    new!([jd, fr], tz, sg)
  end

  class << self; alias_method :new, :civil end

  def self.commercial(*args)
    y, w, d, h, min, s, tz, sg, = parsearg([[-4712, :y, :year, :cwyear],
					    [1, :w, :week, :cweek],
					    [1, :d, :day, :cwday],
					    [0, :h, :hour],
					    [0, :min, :minute],
					    [0, :s, :sec, :second],
					    [0, :of, :offset, :tz, :timezone],
					    [ITALY, :start]], args)

    unless (jd = _valid_commercial?(y, w, d, sg)) &&
	   (fr = _valid_time?(h, min, s))
      raise ArgumentError, 'invalid date'
    end
    if String === tz
      tz = (zone_to_diff(tz) || 0).to_r/86400
    end
    new!([jd, fr], tz, sg)
  end

  def self.weeknum(*args)
    y, w, d, f, h, min, s, tz, sg, = parsearg([[-4712, :y, :year],
					       [0, :w, :week],
					       [1, :d, :day, :wday],
					       [0, :f, :first],
					       [0, :h, :hour],
					       [0, :min, :minute],
					       [0, :s, :sec, :second],
					       [0, :of, :offset, :tz, :timezone],
					       [ITALY, :start]], args)

    unless (jd = _valid_weeknum?(y, w, d, f, sg)) &&
	   (fr = _valid_time?(h, min, s))
      raise ArgumentError, 'invalid date'
    end
    if String === tz
      tz = (zone_to_diff(tz) || 0).to_r/86400
    end
    new!([jd, fr], tz, sg)
  end

  private_class_method :weeknum

  def self.nth_kday(*args)
    y, m, n, k, h, min, s, tz, sg, = parsearg([[-4712, :y, :year],
					       [1, :m, :mon, :month],
					       [1, :n, :nth],
					       [1, :k, :kday, :wday],
					       [0, :h, :hour],
					       [0, :min, :minute],
					       [0, :s, :sec, :second],
					       [0, :of, :offset, :tz, :timezone],
					       [ITALY, :start]], args)

    unless (jd = _valid_nth_kday?(y, m, n, k, sg)) &&
	   (fr = _valid_time?(h, min, s))
      raise ArgumentError, 'invalid date'
    end
    if String === tz
      tz = (zone_to_diff(tz) || 0).to_r/86400
    end
    new!([jd, fr], tz, sg)
  end

#  private_class_method :nth_kday

  def self.new_by_frags(elem, sg)
    elem = rewrite_frags(elem)
    elem = complete_frags(elem)
    unless (jd = valid_date_frags?(elem, sg)) &&
	   (fr = valid_time_frags?(elem))
      raise ArgumentError, 'invalid date'
    end
    sf = (elem[:sec_fraction] || 0)
    fr += sf/86400
    of = (elem[:offset] || 0)
    of = of.to_r/86400
    new!([jd, fr], of, sg)
  end

  private_class_method :new_by_frags

  def self.strptime(*args)
    str, fmt, sg, lc, = parsearg([['-4712-01-01T00:00:00+00:00', :d, :date],
				  ['%FT%T%z', :fmt, :format],
				  [ITALY, :start],
				  [Format::Locale::C, :lc, :locale]], args)
    elem = _strptime(str, fmt, lc)
    new_by_frags(elem, sg)
  end

  def self.parse(*args)
    str, comp, sg, lc, = parsearg([['-4712-01-01T00:00:00+00:00', :d, :date],
				   [false, :comp, :complete],
				   [ITALY, :start],
				   [Format::Locale::C, :lc, :locale]], args)
    elem = _parse(str, comp, lc)
    new_by_frags(elem, sg)
  end

  public :hour, :min, :sec, :sec_fraction, :zone, :offset, :new_offset

  def to_s()
    super() +
      strftime('T%T' +
	       if utc?
		 'Z'
	       else
		 '%:z'
	       end)
  end

end
