# date4/format/strftime.rb: Written by Tadayoshi Funaba 1999-2007
# $Id: strftime.rb,v 1.12 2007-01-07 09:16:58+09 tadf Exp $

class Date4

  def emit(e, f)
    case e
    when Numeric
      sign = %w(+ + -)[e <=> 0]
      e = e.abs
    end

    s = e.to_s

    if f[:s] && f[:p] == '0'
      f[:w] -= 1
    end

    if f[:s] && f[:p] == "\s"
      s[0,0] = sign
    end

    if f[:p] != '-'
      s = s.rjust(f[:w], f[:p])
    end

    if f[:s] && f[:p] != "\s"
      s[0,0] = sign
    end

    s = s.upcase if f[:u]
    s = s.downcase if f[:d]
    s
  end

  def emit_w(e, w, f)
    f[:w] = [f[:w], w].compact.max
    emit(e, f)
  end

  def emit_n(e, w, f)
    f[:p] ||= '0'
    emit_w(e, w, f)
  end

  def emit_sn(e, w, f)
    if e < 0
      w += 1
      f[:s] = true
    end
    emit_n(e, w, f)
  end

  def emit_z(e, w, f)
    w += 1
    f[:s] = true
    emit_n(e, w, f)
  end

  def emit_a(e, w, f)
    f[:p] ||= "\s"
    emit_w(e, w, f)
  end

  def emit_ad(e, w, f)
    if f[:x]
      f[:u] = true
      f[:d] = false
    end
    emit_a(e, w, f)
  end

  def emit_au(e, w, f)
    if f[:x]
      f[:u] = false
      f[:d] = true
    end
    emit_a(e, w, f)
  end

  private :emit, :emit_w, :emit_n, :emit_sn, :emit_z,
	  :emit_a, :emit_ad, :emit_au

  def strftime(*args)
    fmt, lc, = parsearg([['%F', :fmt, :format],
			 [Format::Locale::C, :lc, :locale]], args)
    fmt.gsub(/%([-_0^#]+)?(\d+)?[EO]?(:{1,3}z|.)/m) do |m|
      f = {}
      s, w, c = $1, $2, $3
      if s
	s.scan(/./) do |k|
	  case k
	  when '-'; f[:p] = '-'
	  when '_'; f[:p] = "\s"
	  when '0'; f[:p] = '0'
	  when '^'; f[:u] = true
	  when '#'; f[:x] = true
	  end
	end
      end
      if w
	f[:w] = w.to_i
      end
      case c
      when 'A'; emit_ad(lc.daynames[wday], 0, f)
      when 'a'; emit_ad(lc.abbr_daynames[wday], 0, f)
      when 'B'; emit_ad(lc.monthnames[mon], 0, f)
      when 'b'; emit_ad(lc.abbr_monthnames[mon], 0, f)
      when 'C'; emit_sn((year / 100).floor, 2, f)
      when 'c'; emit_a(strftime(lc.c_fmt, lc), 0, f)
      when 'D'; emit_a(strftime('%m/%d/%y', lc), 0, f)
      when 'd'; emit_n(mday, 2, f)
      when 'e'; emit_a(mday, 2, f)
      when 'F'
	if m == '%F'
	  format('%.4d-%02d-%02d', year, mon, mday) # 4p
	else
	  emit_a(strftime('%Y-%m-%d', lc), 0, f)
	end
      when 'G'; emit_sn(cwyear, 4, f)
      when 'g'; emit_n(cwyear % 100, 2, f)
      when 'H'; emit_n(hour, 2, f)
      when 'h'; emit_ad(strftime('%b', lc), 0, f)
      when 'I'; emit_n((hour % 12).nonzero? || 12, 2, f)
      when 'j'; emit_n(yday, 3, f)
      when 'k'; emit_a(hour, 2, f)
      when 'L'
	emit_n((sec_fraction / (1.to_r/(10**3))).round, 3, f)
      when 'l'; emit_a((hour % 12).nonzero? || 12, 2, f)
      when 'M'; emit_n(min, 2, f)
      when 'm'; emit_n(mon, 2, f)
      when 'N'
	emit_n((sec_fraction / (1.to_r/(10**9))).round, 9, f)
      when 'n'; "\n"
      when 'P'; emit_ad(strftime('%p', lc).downcase, 0, f)
      when 'p'; emit_au(lc.ampmname[if hour < 12 then 0 else 1 end], 0, f)
      when 'Q'
	d = ajd - jd_to_ajd(self.class::UNIXEPOCH, 0)
	s = (d * 86400*10**3).to_i
	emit_sn(s, 1, f)
      when 'R'; emit_a(strftime('%H:%M', lc), 0, f)
      when 'r'; emit_a(strftime(lc.r_fmt, lc), 0, f)
      when 'S'; emit_n(sec, 2, f)
      when 's'
	d = ajd - jd_to_ajd(self.class::UNIXEPOCH, 0)
	s = (d * 86400).to_i
	emit_sn(s, 1, f)
      when 'T'
	if m == '%T'
	  format('%02d:%02d:%02d', hour, min, sec) # 4p
	else
	  emit_a(strftime('%H:%M:%S', lc), 0, f)
	end
      when 't'; "\t"
      when 'U', 'W'
	emit_n(if c == 'U' then wnum0 else wnum1 end, 2, f)
      when 'u'; emit_n(cwday, 1, f)
      when 'V'; emit_n(cweek, 2, f)
      when 'v'; emit_a(strftime('%e-%b-%Y', lc), 0, f)
      when 'w'; emit_n(wday, 1, f)
      when 'X'; emit_a(strftime(lc.X_fmt, lc), 0, f)
      when 'x'; emit_a(strftime(lc.x_fmt, lc), 0, f)
      when 'Y'; emit_sn(year, 4, f)
      when 'y'; emit_n(year % 100, 2, f)
      when 'Z'; emit_au(zone, 0, f)
      when /\A(:{0,3})z/
	t = $1.size
	sign = if offset < 0 then -1 else +1 end
	fr = offset.abs
	hh, fr = fr.divmod(1.to_r/24)
	mm, fr = fr.divmod(1.to_r/1440)
	ss, fr = fr.divmod(1.to_r/86400)
	if t == 3
	  if    ss.nonzero? then t =  2
	  elsif mm.nonzero? then t =  1
	  else                   t = -1
	  end
	end
	case t
	when -1
	  tail = []
	  sep = ''
	when 0
	  f[:w] -= 2 if f[:w]
	  tail = ['%02d' % mm]
	  sep = ''
	when 1
	  f[:w] -= 3 if f[:w]
	  tail = ['%02d' % mm]
	  sep = ':'
	when 2
	  f[:w] -= 6 if f[:w]
	  tail = ['%02d' % mm, '%02d' % ss]
	  sep = ':'
	end
	([emit_z(sign * hh, 2, f)] + tail).join(sep)
      when '%'; emit_a('%', 0, f)
      when '+'; emit_a(strftime(lc.date_fmt, lc), 0, f)
      else
	c
      end
    end
  end

# alias_method :format, :strftime

  def asctime() strftime('%c') end

  alias_method :ctime, :asctime

  def iso8601() strftime('%F') end

  def rfc3339() iso8601 end

  def rfc2822() strftime('%a, %-d %b %Y %T %z') end

  alias_method :rfc822, :rfc2822

  def jisx0301
    if jd < 2405160
      iso8601
    else
      case jd
      when 2405160...2419614
	g = 'M%02d' % (year - 1867)
      when 2419614...2424875
	g = 'T%02d' % (year - 1911)
      when 2424875...2447535
	g = 'S%02d' % (year - 1925)
      else
	g = 'H%02d' % (year - 1988)
      end
      g + strftime('.%m.%d')
    end
  end

=begin
  def beat(n=0)
    i, f = (new_offset(1.to_r/24).day_fraction * 1000).divmod(1)
    ('@%03d' % i) +
      if n < 1
	''
      else
	'.%0*d' % [n, (f / (1.to_r/(10**n))).round]
      end
  end
=end

end

class DateTime4 < Date4

  def strftime(*args)
    fmt, lc, = parsearg([['%FT%T%:z', :fmt, :format],
			 [Format::Locale::C, :lc, :locale]], args)
    super(fmt, lc)
  end

  def iso8601_timediv(n)
    strftime('T%T' +
	     if n < 1
	       ''
	     else
	       '.%0*d' % [n, (sec_fraction / (1.to_r/(10**n))).round]
	     end +
	     if utc? then '%Z' else '%:z' end)
  end

  private :iso8601_timediv

  def iso8601(n=0)
    super() + iso8601_timediv(n)
  end

  def rfc3339(n=0) iso8601(n) end

  def jisx0301(n=0)
    super() + iso8601_timediv(n)
  end

end
