# date4/timezone/posix.rb: Written by Tadayoshi Funaba 2006,2007
# $Id: posix.rb,v 1.3 2007-04-07 08:30:33+09 tadf Exp $

require 'date4/holiday'
require 'date4/timezone/posix/parser'

class Date4

  module TimeZone

    class POSIX

      def initialize(spec)
	t = begin (@@pa ||= Parser.new).parse(@spec = spec)
          rescue Racc::ParseError
            raise ArgumentError, 'syntax error'
          end

	stdof, dstof, rule, = t
	@std, @ostd, = stdof
	@dst, @odst, = dstof
	sdt, edt, = rule
	@sd, @st, = sdt
	@ed, @et, = edt

	if @dst && !@sd
	  @sd, @st, @ed, @et, = US::Eastern.rule
	end

	@odst ||= @ostd - 3600
	@ostd = -@ostd
	@odst = -@odst
	@ostd = @ostd.to_r/86400
	@odst = @odst.to_r/86400
	@st ||= 7200
	@et ||= 7200
	@st = @st.to_r/86400
	@et = @et.to_r/86400
      end

      def rule() [@sd, @st, @ed, @et] end

      protected :rule

      def fix(date, year)
	case date[0]
	when :J1
	  d = date[1]
	  if d < 60
	    d -= 1
	    DateTime4.new(year, 1, 1) + d
	  else
	    d -= 60
	    DateTime4.new(year, 3, 1) + d
	  end
	when :J0
	  d = date[1]
	  DateTime4.new(year, 1, 1) + d
	when :M
	  m = date[1]
	  n = date[2]
	  k = date[3]
	  if n == 5
	    n = -1
	  end
	  DateTime4.nth_kday(year, m, n, k)
	end
      end

      private :fix

      def lookup(ajd, adj)
	unless @dst
	  dst = false
	else
	  as_loc = DateTime4.new!(ajd)
	  if adj
	    list = [as_loc]
	  else
	    as_dst = as_loc + @odst
	    as_std = as_loc + @ostd
	    list = as_dst, as_std
	  end
	  diff = @odst - @ostd
	  catch :exit do
	    list.collect{|d| d.year}.uniq.each do |y|
	      dst_start = fix(@sd, y) + (@st + diff)
	      dst_end   = fix(@ed, y) + (@et - diff)
	      southern = dst_start > dst_end
	      if southern
		dst_start, dst_end = dst_end, dst_start
	      end
	      if adj
		l, r = as_loc, as_loc
	      else
		l, r = as_dst, as_std
		if southern
		  l, r = r, l
		end
	      end
	      dst = l >= dst_start && r < dst_end
	      if southern
		dst = !dst
	      end
	      if dst
		throw :exit
	      end
	    end
	  end
	end
	zone = if dst then @dst else @std end
	offset = if dst then @odst else @ostd end
	{:zone=>zone, :of=>offset, :utc=>false, :dst=>dst}
      end

      def inspect() format('#<%s: %s>', self.class, @spec) end
      def to_s() @spec.dup end

      module AU
	Eastern = POSIX.new('EST-10EST,M10.5.0/2,M3.5.0/3')
      end

      module JP
	Central = POSIX.new('JST-9')
	Standard = Central
      end

      module TW
	Standard = POSIX.new('CST-8')
      end

      CN = TW

      module TH
	Standard = POSIX.new('ICT-7')
      end

      module PK
	Standard = POSIX.new('PKT-5')
      end

      module RU
	Moscow = POSIX.new('MSK-3MSD,M3.5.0/2,M10.5.0/3')
      end

      module EG
	Standard = POSIX.new('EST-3')
      end

      module EU
	Eastern = POSIX.new('EET-2EEST,M3.5.0/3,M10.5.0/4')
	Central = POSIX.new('CET-1CEST,M3.5.0/2,M10.5.0/3')
	Western = POSIX.new('WET0WEST,M3.5.0/1,M10.5.0/2')
      end

      module UK
	British = POSIX.new('GMT0BST,M3.5.0/1,M10.5.0/2')
	Irish = POSIX.new('GMT0IST,M3.5.0/1,M10.5.0/2')
      end

      module Etc
	UTC = POSIX.new('UTC0')
	UCT = UTC
      end

      module US
	Atlantic = POSIX.new('AST4ADT,M3.2.0,M11.1.0')
	Eastern = POSIX.new('EST5EDT,M3.2.0,M11.1.0')
	Central = POSIX.new('CST6CDT,M3.2.0,M11.1.0')
	Mountain = POSIX.new('MST7MDT,M3.2.0,M11.1.0')
	Pacific = POSIX.new('PST8PDT,M3.2.0,M11.1.0')
	Alaska = POSIX.new('AKST9AKDT,M3.2.0,M11.1.0')
	Hawaii = POSIX.new('HST10')
      end

    end

  end

end
